#!/usr/bin/perl -w

# Tool to upload Magellan 315/320 POI data file to the unit
# Copytinght 2002 Cvetan Ivanov <zezo@inet.bg>
# version 0.11
# Args: filename.bin
#
# This program is tested only on my unit and works for me.
# But beware, it may destroy your unit.
# Use at your own risk.
# 
# Test 98 is entered by pressing 
# menu - right arrow - left arrow - right arrow
# entering 98, and pressing enter. 
# It's matter of 3 more lines of code to do
# it with a command, but I prefer it this way. 
# You should know what you're doing.
#
# Be careful _not_ to activate test 99, which is firmware upload
# and most certainly _will_ kill your unit

use strict;

use IO::File;
use IO::Stty;

my $tty='/dev/ttyS1';
my $debug = 0;

my $state='init';
my $DEVICE;
my $infile;
my $dbsize;

sub HELLO_CMD() 	{0xAA};
sub OK_CMD() 		{0x55};
sub ERASE_CMD() 	{0x6A};
sub WAIT_CMD()		{0x77};
sub COLDSTART_CMD() 	{0x69};
sub ERASE_OK() 		{0x8F};
sub DEBUG_CMD() 	{0x8e};

sub maxbase() { 1441790 };

sub initio()
{
	die "Cannot set serial speed" if system("stty", "-F", "$tty", "115200");
        $DEVICE=new IO::File;
	die "Cannot open $tty" unless $DEVICE->open("+<$tty");
        $DEVICE->autoflush(1);
        IO::Stty::stty($DEVICE, 'raw');
	IO::Stty::stty($DEVICE, '-echo','-hupcl', 'clocal', '-cstopb', 'pass8');
}

sub checksum($) 
{
	my $data = shift;
	my $i = 1;
	my $sum = 0;
	
	for($i=0; $i < length($data); $i+=2) {
		$sum ^= unpack("n", substr($data, $i, 2));
	}
	return $sum;
}

sub build_record($)
{
	my $data = shift;
	return '[' . pack("n*", length($data)/2, unpack("n*",$data), checksum($data)) . ']';
}

sub myread($) 
{
	my $count = shift;
        my $result=eval {
                my $r;
		my $n;
                local $SIG{ALRM} = sub { die "timeout" };
                alarm 10;
		$n = $DEVICE->sysread($r,$count);
                printf STDERR ("Got %d bytes\n", $n) if $debug;
                alarm 0;
                return $r;
        };
        die "Timeout communicating with unit" if $@;
	return $result; 
}

sub waitfor($)
{
	my $expected = shift;
	my $r;
	my $off;
	while (1) {
		$r = myread(1);
		$r = unpack("C", $r);
		printf STDERR (" Expecting: %x read: %x\n", $expected, $r) if $debug;
		if ($r == 0x81) { # bad crc
			$DEVICE->sysread($r,2);
			return 0;
		}
		if ($r == 0x88) { # short record
			$DEVICE->sysread($r,4);
			return 0;
		}
		return 0 if ($r == 0x85); #incomplete record

		die "Error: Record too long" 		if ($r == 0x82);
		die "Cannot program/erase" 		if ($r == 0x84);
		die "Erase command at wrong time" 	if ($r == 0x83);
		
		next  if ($state ne 'init' && $r == 0x77); #resync

		if ($state eq 'erasing' && $r == ERASE_OK) {
			print ".";
			next;
		}

		if ($state eq 'programming' && $r == DEBUG_CMD) {	
			$DEVICE->sysread($off,2);
			$DEVICE->sysread($r,1);
			$DEVICE->sysread($off,2,2);
			printf("%07u (%d%%)\r", unpack("N", $off) - 0xa0000, 
					(unpack("N", $off) - 0xa0000)/$dbsize*100);
			next;
		}
			
		if ($state eq 'waitack' && $r == HELLO_CMD) {
			printf STDERR ("Ignoring: %x\n", $r) if $debug;
			next;
		}

		if ($r == $expected) 
		{
			return 1;
		}
		last; 		
	}
	die ("Protocol error\n");
}

sub connect_gps()
{
	print "Activate test 98 and press Enter\n";
	STDIN->flush();
	<>;
	print "Waiting for link\n";
	waitfor(HELLO_CMD);
	$state='waitack';
	$DEVICE->print(pack("C", OK_CMD));
	waitfor(OK_CMD);
	print "Connected\n";
	$state='connected';
}
		
sub erase_flash()
{
	print "Erasing ";
	$state='erasing';
	$DEVICE->print(pack("C", ERASE_CMD));
	waitfor(OK_CMD);
	$DEVICE->print(build_record(pack("NN",0,$dbsize)));
	waitfor(OK_CMD);
	waitfor(OK_CMD);
	print "\nErased.\n";
}

sub upload_file()
{
	my $count;
	my $buff;
	my $offset = 0;
	printf("Programming %07u bytes.\n", $dbsize);
	$state = 'programming';
	while ($count = sysread(IN, $buff, 2048)) {
		$DEVICE->print(build_record(pack("N",$offset) . $buff));
		while (waitfor(OK_CMD) == 0) { # retry on some errors
			print STDERR ("Got recoverable error. Retrying\n") if $debug;
			$DEVICE->print(build_record(pack("N",$offset) . $buff));
		}
		waitfor(OK_CMD);
		$offset += $count;
	}
        $DEVICE->print(build_record(""));
        waitfor(OK_CMD);
	print("Programmed OK\nRestarting\n");
}
	
$infile = shift;
die "No input file" unless $infile;
open (IN, "<$infile") or die "Cannot open $infile";
STDOUT->autoflush(1);
STDERR->autoflush(1);

$dbsize=(stat(IN))[7];
die "Input file too big" if  $dbsize > maxbase;

initio;
connect_gps;
erase_flash;
upload_file;
